/* glwindow.c
 * 
 * Kevin P. Smith  10/28/90
 *
 */
#include "copyright2.h"

#include <stdio.h>
#include <gl.h>
#include <device.h>
#include "Wlib.h"

#define FONTS 4
#define COLORS 7

#undef WHITE
#undef BLACK
#undef RED
#undef GREEN
#undef YELLOW
#undef CYAN

#define WHITE   0
#define BLACK   1
#define RED     2
#define GREEN   3
#define YELLOW  4
#define CYAN    5
#define GREY    6

static int zero=0;
static int one=1;
static int two=2;
static int three=3;

int W_FastClear=1;
W_Font W_BigFont= (W_Font) &zero, W_RegularFont= (W_Font) &one;
W_Font W_HighlightFont= (W_Font) &two, W_UnderlineFont= (W_Font) &three;
W_Color W_White=WHITE, W_Black=BLACK, W_Red=RED, W_Green=GREEN;
W_Color W_Yellow=YELLOW, W_Cyan=CYAN, W_Grey=GREY;
int W_Textwidth, W_Textheight;
static int lastFontLoaded;

static int lastfont = 0;
static W_Window lastwindow = 0;

Fontchar fonts[128];

unsigned short font6x10[] = {
#include "6x10.font.h"
};

unsigned short font6x10i[] = {
#include "6x10i.font.h"
};

unsigned short font6x10b[] = {
#include "6x10b.font.h"
};

unsigned short fontsbdr40sx[] = {
#include "sbdr40sx.font.h"
};

static W_Event W_myevent;
static int W_isEvent=0;
static int firstwindow=1;

struct window *focus=NULL;

struct icon {
    struct window *window;
    unsigned char *bits;
    int width, height;
    int fontnum, charnum;
};

struct colors {
    char *name;
    int pixelvalue;
    short r,g,b;
};

#define WIN_GRAPH       1
#define WIN_TEXT        2
#define WIN_MENU        3
#define WIN_SCROLL      4

#define FLIPY(y) (1023-(y))

struct window {
    long window;
    int type;
    char *data;
    int x,y;
    int width,height;
    int offx, offy;
    int winwidth, winheight;
    char *name;
    int lb,rb,tb,bb;	/* border sizes for all sides of the window */
    int border;
    int cursor;
    W_Color lastcolor;
    int domap;		/* Do I map when my parent maps? */
    W_Color color;
    W_Window parent;
    struct windowlist *children;
    unsigned short *tile;
};

struct stringList {
    char *string;
    W_Color color;
    struct stringList *next;
};

struct menuItem {
    char *string;
    W_Color color;
};

struct colors colortable[COLORS] = {
    { "white"      , 1031, 0xff, 0xff, 0xff },
    { "black"      , 1024, 0x00, 0x00, 0x00 },
    { "red"        , 1025, 0xff, 0x5f, 0x5f },
    { "green"      , 1030, 0x5f, 0xff, 0x5f },
    { "yellow"     , 1028, 0xff, 0xff, 0x5f },
    { "cyan"       , 1026, 0x5f, 0xff, 0xff },
    { "light grey" , 1027, 0xa0, 0xa0, 0xa0 },
};

struct windowlist {
    struct window *window;
    struct windowlist *next;
};

#define HASHSIZE 29
#define hash(x) (x % HASHSIZE)

struct windowlist *hashtable[HASHSIZE];
struct window *newWindow();
struct window *findWindow();
char *malloc();
unsigned short *x11tocursor();
unsigned short *makeTile();
char *strdup();
char *getdefault();

#define W_Void2Window(win) ((struct window *) (win))
#define W_Window2Void(window) ((W_Window) window)
#define W_Void2Icon(bit) ((struct icon *) (bit))
#define W_Icon2Void(bit) ((W_Icon) (bit))
#define fontNum(font) (*((int *) font)+1)

#define WIN_EDGE	5
#define MENU_PAD	4
#define MENU_BAR	4

static int v[2];
#define V2(x,y) {v[0]=x; v[1]=y; v2i(v);}

void
W_Initialize(display)
char *display;
{
    W_Textheight=10;
    W_Textwidth=6;

    noport();
    foreground();
    winopen("init");
    realInitialize();
}

realInitialize()
{
    int i;
    
    for (i=0; i<COLORS; i++) {
	mapcolor(colortable[i].pixelvalue, 
	    colortable[i].r,
	    colortable[i].g,
	    colortable[i].b);
    }
    qdevice(LEFTMOUSE);
    qdevice(MOUSEX);
    qdevice(MOUSEY);
    qdevice(MIDDLEMOUSE);
    qdevice(RIGHTMOUSE);
    qdevice(KEYBD);
    qdevice(REDRAW);
    qdevice(INPUTCHANGE);
    for (i=0; i<128; i++) {
	fonts[i].w=6;
	fonts[i].h=10;
	fonts[i].width=6;
	fonts[i].xoff=0;
	fonts[i].yoff=0;
	fonts[i].offset=10*i;
    }
    defrasterfont(fontNum(W_RegularFont),10,128,fonts,10*128,font6x10);
    defrasterfont(fontNum(W_HighlightFont),10,128,fonts,10*128,font6x10b);
    defrasterfont(fontNum(W_UnderlineFont),10,128,fonts,10*128,font6x10i);
    for (i=0; i<128; i++) {
	fonts[i].w=32;
	fonts[i].h=52;
	fonts[i].width=32;
	fonts[i].xoff=0;
	fonts[i].yoff=0;
	fonts[i].offset=104*i;
    }
    defrasterfont(fontNum(W_BigFont),52,128,fonts,104*128,fontsbdr40sx);
}

void
W_DefineCursor(window,width,height,bits,mask,xhot,yhot)
W_Window window;
int width,height,xhot,yhot;
unsigned char *bits, *mask;
{
    unsigned short *data;
    int i;
    static int cursnum=1;
    struct window *win;
    int cnum;
    struct windowlist *wl;

    if (height != 16 || width != 16) {
	fprintf(stderr, "Sorry, but cursor must be 16 by 16\n");
	return;
    }

    win=W_Void2Window(window);
    data=x11tocursor(bits);

    if (win->cursor) cnum=win->cursor;
    else cnum=cursnum++;

    defcursor(cnum, data);
    curorigin(cnum, xhot, (height-yhot-1));
    win->cursor=cnum;

    if (win->window) {
	setWindow(window);
	setcursor(cnum);
    }
    wl=win->children;
    while (wl) {
	if (wl->window->cursor == 0) {
	    W_DefineCursor(wl->window,width,height,bits,mask,xhot,yhot);
	}
	wl=wl->next;
    }
    free((char *) data);
}

W_Window
W_MakeWindow(name, x, y, width, height, parent, border, color)
char *name;
int x,y,width,height;
W_Window parent;
int border;
W_Color color;
{
    struct window *newwin;

    checkGeometry(name,&x,&y,&width,&height);
    checkParent(name, &parent);

    newwin=newWindow(name,WIN_GRAPH,x,y,width,height,border,color,parent,
	width+2*border, height+2*border);

    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
    return(W_Window2Void(newwin));
}

void
W_ChangeBorder(window, color)
W_Window window;
W_Color color;
{
    struct window *win;

    win=W_Void2Window(window);
    win->color=color;
    drawBorder(win);
}

int
W_IsMapped(window)
W_Window window;
{
    struct window *win;

    win=W_Void2Window(window);
    return(win->domap);
}

void
W_ClearArea(window, x, y, width, height, color)
W_Window window;
int x,y;
int width,height;
W_Color color;
{
    struct window *win;

    win=W_Void2Window(window);
    if (win->window == 0) return;
    setWindow(window);
    setColor(color);

    switch(win->type) {
    case WIN_GRAPH:
	rectfi(x,y+1,x+width-1,y+height);
	break;
    default:
	rectfi(x*W_Textwidth+WIN_EDGE,
	    y*W_Textheight+MENU_PAD+height*W_Textheight,
	    x*W_Textwidth+WIN_EDGE+W_Textwidth*width-1,
	    y*W_Textheight+MENU_PAD+1);
	break;
    }
}

void
W_ClearWindow(window)
W_Window window;
{
    struct window *win;

    win=W_Void2Window(window);
    if (win->window == 0) return;
    setWindow(window);
    setColor(W_Black);

    if (win->tile == NULL) clear();
    else drawTile(win);
}

void
W_UnmapWindow(window)
W_Window window;
{
    struct window *win;
    struct windowlist *children;

    win=W_Void2Window(window);

    children=win->children;
    while (children) {
	if (children->window->domap) {
	    W_UnmapWindow(W_Window2Void(children->window));
	    children->window->domap=1;
	}
	children=children->next;
    }

    if (win->window == 0) return;

    setWindow(window);
    winclose(win->window);

    removeFromHash(win);
    win->domap=0;
    win->window=0;
}

W_Icon
W_StoreBitmap(width,height,bits,window)
int width,height;
unsigned char *bits;
W_Window window;
{
    struct icon *newicon;

    newicon=(struct icon *) malloc(sizeof(struct icon));
    newicon->window=W_Void2Window(window);
    newicon->height=height;
    newicon->width=width;
    newicon->bits=bits;
    loadNewBitmap(newicon);
    return(W_Icon2Void(newicon));
}

void
W_WriteBitmap(x,y,bitmap,color)
int x,y;
W_Icon bitmap;
W_Color color;
{
    struct icon *pic;
    struct window *win;
    char str[2];

    pic=W_Void2Icon(bitmap);
    win=pic->window;
    setWindow(W_Window2Void(win));
    setColor(color);
    if (lastFontLoaded<pic->fontnum) storeNewFont();

    cmov2i(x,y+pic->height);
    setFont(pic->fontnum);
    str[1]=0;
    str[0]=pic->charnum;
    charstr(str);
}

int
W_EventsPending()
{
    if (W_isEvent) return(1);
    while (qtest()) {
	if (W_SpNextEvent(&W_myevent)) {
	    W_isEvent=1;
	    return(1);
	}
    }
    return(0);
}

void
W_NextEvent(event) 
W_Event *event;
{
    if (W_isEvent) {
	*event=W_myevent;
	W_isEvent=0;
	return;
    }
    while (W_SpNextEvent(event)==0);
}

int
W_SpNextEvent(event)
W_Event *event;
{
    short dev, val;
    static int x,y;
    int legal;
    struct window *win;
    struct window *newfocus;

    legal=0;
    if (focus) {
	if (focus->type == WIN_TEXT || focus->type == WIN_SCROLL) legal=0;
	event->Window=W_Window2Void(focus);
	event->x=x-focus->x-focus->lb;
	event->y=FLIPY(focus->y)-y-focus->tb;
	legal=1;
	if (focus->type == WIN_MENU) {
	    if (event->y % (W_Textheight + MENU_PAD*2+MENU_BAR) >=
		W_Textheight+MENU_PAD*2) legal=0;
	    event->y=event->y/(W_Textheight+MENU_PAD*2+MENU_BAR);
	    if (event->y<0 || event->y>=focus->height) legal=0;
	} else if (!(event->x>=0 && event->y>=0 && event->x<focus->width &&
	    event->y<focus->height)) legal=0;
    } else legal=0;
    dev=qread(&val);
    switch(dev) {
    case MOUSEX:
	x=val;
	break;
    case MOUSEY:
	y=val;
	break;
    case REDRAW:
	event->type=W_EV_EXPOSE;
	win=findWindow(val);
	event->Window=W_Window2Void(win);
	if (win) {
	    drawBorder(win);
	    drawTile(win);
	    if (win->type == WIN_SCROLL) {
		redrawScrolling(win);
		return(0);
	    }
	    if (win->type == WIN_MENU) {
		redrawMenu(win);
		return(0);
	    }
	    return(1);
	}
	break;
    case INPUTCHANGE:
	newfocus=findWindow(val);
	/* I refuse to focus on a text or scroll window */
	if (newfocus==NULL || 
	    newfocus->type == WIN_TEXT || 
	    newfocus->type == WIN_SCROLL) break;
	focus=newfocus;
	break;
    case LEFTMOUSE:
	if (val==0) break;
	event->type=W_EV_BUTTON;
	event->key=W_LBUTTON;
	return(legal);
    case MIDDLEMOUSE:
	if (val==0) break;
	event->type=W_EV_BUTTON;
	event->key=W_MBUTTON;
	return(legal);
    case RIGHTMOUSE:
	if (val==0) break;
	event->type=W_EV_BUTTON;
	event->key=W_RBUTTON;
	return(legal);
    case KEYBD:
	event->type=W_EV_KEY;
	event->key=val;
	return(legal);
    default:
	break;
    }
    return(0);
}

void
W_MapWindow(window)
W_Window window;
{
    struct window *win;
    struct window *parent;
    struct windowlist *children;

    win=W_Void2Window(window);
    win->domap=1;
    if (win->window != 0) return;

    parent=W_Void2Window(win->parent);
    if (parent && parent->window == 0) return;

    noborder();
    foreground();
    prefposition(win->x, win->x+win->winwidth-1, 
	FLIPY(win->y), FLIPY(win->y+win->winheight-1));
    win->window=winopen(win->name);
    lastfont=0;
    qenter(REDRAW, win->window);
    addToHash(win);
    if (firstwindow) {
	focus=win;
	firstwindow=0;
    }
    if (win->cursor) {
	setcursor(win->cursor);
    }
    shapewindow(win);
    drawBorder(win);

    children=win->children;
    while (children) {
	if (children->window->domap) {
	    W_MapWindow(W_Window2Void(children->window));
	}
	children=children->next;
    }
}

void
W_MakeLine(window, x0, y0, x1, y1, color)
W_Window window;
int x0,y0,x1,y1;
W_Color color;
{
    struct window *win;

    win=W_Void2Window(window);
    if (win->window == 0) return;
    setWindow(window);
    setColor(color);
    bgnline();
    V2(x0,y0);
    V2(x1,y1);
    endline();
}

int 
W_WindowWidth(window)
W_Window window;
{
    struct window *win;

    win=W_Void2Window(window);
    return(win->winwidth-win->border*2);
}

int
W_WindowHeight(window)
W_Window window;
{
    struct window *win;

    win=W_Void2Window(window);
    return(win->height-win->border*2);
}

int
W_Socket()
{
    /* Sorry.  No socket. */
    return(-1);
}

void
W_Beep()
{
    /* hehehehehe.  How's this? */
    fprintf(stderr, "");
}

void
W_SetIconWindow(main, icon)
W_Window main;
W_Window icon;
{
    /* Dream on! */
}

void
W_DestroyWindow(window)
W_Window window;
{
    struct window *win;
    struct windowlist *wl;
    struct windowlist *temp;

    win=W_Void2Window(window);

    wl=win->children;
    while (wl) {
	W_DestroyWindow(wl->window);
	temp=wl;
	wl=wl->next;
	free((char *) temp);
    }
    win->children=NULL;

    W_UnmapWindow(window);

    if (win->parent) {
	deleteChild(W_Window2Void(win->parent),W_Window2Void(win));
    }

    if (win->name) free(win->name);
    free((char *) win);
}

W_Window
W_MakeMenu(name,x,y,width,height,parent,border)
char *name;
int x,y,width,height;
W_Window parent;
int border;
{
    struct window *newwin;
    struct menuItem *items;
    int i;

    checkGeometry(name,&x,&y,&width,&height);
    checkParent(name,&parent);
    newwin=newWindow(name,WIN_MENU,x,y,width,height,border,W_White,parent,
	width*W_Textwidth+WIN_EDGE*2+border*2,
	height*(W_Textheight+MENU_PAD*2)+(height-1)*MENU_BAR+border*2);

    items=(struct menuItem *) malloc(height*sizeof(struct menuItem));
    for (i=0; i<height; i++) {
	items[i].string=NULL;
	items[i].color=W_White;
    }
    newwin->data=(char *) items;

    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
    return(W_Window2Void(newwin));
}

W_Window
W_MakeScrollingWindow(name,x,y,width,height,parent,border)
char *name;
int x,y,width,height;
W_Window parent;
int border;
{
    struct window *newwin;

    checkGeometry(name,&x,&y,&width,&height);
    checkParent(name,&parent);
    newwin=newWindow(name,WIN_SCROLL,x,y,width,height,border,W_White,parent,
	width*W_Textwidth+WIN_EDGE*2+border*2,
	height*W_Textheight+MENU_PAD*2+border*2);

    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
    return(W_Window2Void(newwin));
}

W_Window
W_MakeTextWindow(name,x,y,width,height,parent,border)
char *name;
int x,y,width,height;
W_Window parent;
int border;
{
    struct window *newwin;

    checkGeometry(name,&x,&y,&width,&height);
    checkParent(name,&parent);
    newwin=newWindow(name,WIN_TEXT,x,y,width,height,border,W_White,parent,
	width*W_Textwidth+WIN_EDGE*2+border*2,
	height*W_Textheight+MENU_PAD*2+border*2);

    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
    return(W_Window2Void(newwin));
}

void
W_UnTileWindow(window)
W_Window window;
{
    struct window *win;

    win=W_Void2Window(window);
    if (win->tile) {
	free((char *) win->tile);
	win->tile=NULL;
    }
    if (win->window) qenter(REDRAW, win->window);
}

void
W_TileWindow(window, bit)
W_Window window;
W_Icon bit;
{
    struct window *win;

    win=W_Void2Window(window);
    if (win->tile) W_UnTileWindow(window);

    win->tile=makeTile(bit,win->winwidth-win->lb-win->rb,
	win->winheight-win->tb-win->bb);
    drawTile(win);
}

void
W_MaskText(window,x,y,color,str,len,font)
W_Window window;
int x,y,len;
W_Color color;
char *str;
W_Font font;
{
    struct window *win;
    char ch;
    int height;

    if (font==W_BigFont) {
	height=52;
    } else {
	height=10;
    }

    win=W_Void2Window(window);
    if (win->window == 0) return;

    setWindow(window);
    setColor(color);
    setFont(fontNum(font));
    cmov2i(x,y+height);
    ch= *(str+len);
    *(str+len)=0;
    charstr(str);
    *(str+len)=ch;
}

void
W_WriteText(window,x,y,color,str,len,font)
W_Window window;
int x,y,len;
W_Color color;
char *str;
W_Font font;
{
    struct window *win;
    char ch;
    int width,height;

    win=W_Void2Window(window);
    if ((win->type==WIN_GRAPH || win->type==WIN_TEXT) && 
	win->window==0) return;
    ch= *(str+len);
    *(str+len)=0;

    if (font==W_BigFont) {
	height=52;
	width=32*len;
    } else {
	height=10;
	width=6*len;
    }

    switch(win->type) {
    case WIN_GRAPH:
	setWindow(window);
	setFont(fontNum(font));
	setColor(W_Black);
	rectfi(x,y+height,x+width-1,y+1);
	retile(win,x,y,x+width-1,y+height-1);
	setColor(color);
	cmov2i(x,y+height);
	charstr(str);
	break;
    case WIN_SCROLL:
	AddToScrolling(win,color,str);
	redrawScrolling(win);
	break;
    case WIN_MENU:
	changeMenuItem(win,y,str,color);
	break;
    case WIN_TEXT:
    default:
	setWindow(window);
	setFont(fontNum(font));
	setColor(W_Black);
	rectfi(x*W_Textwidth+WIN_EDGE,y*W_Textheight+MENU_PAD+height,
	    x*W_Textwidth+WIN_EDGE+width-1,y*W_Textheight+MENU_PAD+1);
	setColor(color);
	cmov2i(x*W_Textwidth+WIN_EDGE,y*W_Textheight+MENU_PAD+height);
	charstr(str);
	break;
    }
    *(str+len)=ch;
}

/* Add a string to the string list of the scrolling window.
 */
AddToScrolling(win, color, str)
struct window *win;
W_Color color;
char *str;
{
    struct stringList **strings;
    char *newstring;
    int count;

    strings= (struct stringList **) &(win->data);
    count=0;
    while ((*strings)!=NULL) {
        count++;
        strings= &((*strings)->next);
    }
    (*strings)=(struct stringList *) malloc(sizeof(struct stringList));
    (*strings)->next=NULL;
    (*strings)->color=color;
    newstring=strdup(str);
    (*strings)->string=newstring;
    if (count >= 100) {         /* Arbitrary large size. */
        struct stringList *temp;
        temp=(struct stringList *) win->data;
        free(temp->string);
        temp=temp->next;
        free((char *) win->data);
        win->data=(char *) temp;
    }
}

redrawScrolling(win)
struct window *win;
{
    int count;
    struct stringList *list;
    int y;

    if (win->window == 0) return;
    setWindow(W_Window2Void(win));
    setColor(W_Black);
    setFont(fontNum(W_RegularFont));
    clear();
    count=0;
    list = (struct stringList *) win->data;
    while (list != NULL) {
        list=list->next;
        count++;
    }
    list = (struct stringList *) win->data;
    while (count > win->height) {
        list=list->next;
        count--;
    }
    y=(win->height-count+1)*W_Textheight;
    if (count==0) return;
    while (list != NULL) {
	setColor(list->color);
	cmov2i(WIN_EDGE, MENU_PAD+y);
	charstr(list->string);
        list=list->next;
        y=y+W_Textheight;
    }
}

redrawMenu(win)
struct window *win;
{
    int count;

    if (win->window == 0) return;
    setWindow(W_Window2Void(win));
    setColor(W_White);

    for (count=1; count<win->height; count++) {
	rectfi(0, count*(W_Textheight+MENU_PAD*2)+(count-1)*MENU_BAR,
	    win->width*W_Textwidth+WIN_EDGE*2, 
	    count*(W_Textheight+MENU_PAD*2)+(count)*MENU_BAR);
    }
    for (count=0; count<win->height; count++) {
        redrawMenuItem(win,count);
    }
}

redrawMenuItem(win, n)
struct window *win;
int n;
{
    struct menuItem *items;

    if (win->window == 0) return;
    setWindow(W_Window2Void(win));
    setColor(W_Black);
    setFont(fontNum(W_RegularFont));

    items=(struct menuItem *) win->data;
    rectfi(WIN_EDGE, n*(W_Textheight+MENU_PAD*2+MENU_BAR)+MENU_PAD,
	WIN_EDGE+win->width*W_Textwidth,
	n*(W_Textheight+MENU_PAD*2+MENU_BAR)+MENU_PAD+W_Textheight);
    if (items[n].string) {
	setColor(items[n].color);
	cmov2i(WIN_EDGE, 
	    n*(W_Textheight+MENU_PAD*2+MENU_BAR)+MENU_PAD+W_Textheight);
	charstr(items[n].string);
    }
}

changeMenuItem(win, n, str, color)
struct window *win;
int n;
char *str;
W_Color color;
{
    struct menuItem *items;
    char *news;

    items=(struct menuItem *) win->data;
    if (items[n].string) {
        free(items[n].string);
    }
    news=strdup(str);
    items[n].string=news;
    items[n].color=color;
    redrawMenuItem(win, n);
}


struct window *
newWindow(name,type,x,y,width,height,border,color,parent,winwidth,winheight)
char *name;
int type,x,y,width,height,border,winwidth,winheight;
W_Color color;
W_Window parent;
{
    struct window *newwin;
    struct window *myparent;

    newwin=(struct window *) malloc(sizeof(struct window));
    myparent=W_Void2Window(parent);
    newwin->lb=newwin->rb=newwin->tb=newwin->bb=border;
    newwin->cursor=0;
    newwin->winwidth=winwidth;
    newwin->winheight=winheight;
    newwin->border=border;
    newwin->offx=0;
    newwin->offy=0;
    if (myparent) {
	addChild(parent, W_Window2Void(newwin));
	newwin->cursor=myparent->cursor;
	if (x<0) {
	    newwin->lb += x;
	    newwin->winwidth += x;
	    if (-x > border) {
		newwin->offx= -border-x;
	    }
	    if (newwin->lb < 0) newwin->lb=0;
	    x=0;
	}
	if (y<0) {
	    newwin->tb += y;
	    newwin->winheight += y;
	    if (-y > border) {
		newwin->offy= -border-y;
	    }
	    if (newwin->tb < 0) newwin->tb=0;
	    y=0;
	}
	if (newwin->winwidth+x>myparent->winwidth-myparent->rb-myparent->lb) {
	    int diff;

	    diff=newwin->winwidth+x-myparent->winwidth+myparent->rb+myparent->lb;

	    newwin->rb -= diff;
	    newwin->winwidth -= diff;
	    if (newwin->rb < 0) {
		newwin->rb=0;
	    }
	}
	if (newwin->winheight+y>myparent->winheight-myparent->bb-myparent->tb) {
	    int diff;

	    diff=newwin->winheight+y-myparent->winheight+myparent->bb+myparent->tb;
	    newwin->bb -= diff;
	    newwin->winheight -= diff;
	    if (newwin->bb < 0) {
		newwin->bb=0;
	    }
	}
	x += myparent->x + myparent->lb;
	y += myparent->y + myparent->tb;
    }
    newwin->name=strdup(name);
    newwin->window=0;
    newwin->type=type;
    newwin->x=x;
    newwin->y=y;
    newwin->width=width;
    newwin->height=height;
    newwin->border=border;
    newwin->color=color;
    newwin->parent=parent;
    newwin->data=NULL;
    newwin->children=NULL;
    newwin->tile=NULL;
    newwin->domap=0;
    return(newwin);
}

addChild(parent, child)
W_Window parent, child;
{
    struct window *win;
    struct windowlist *wl;

    wl=(struct windowlist *) malloc(sizeof(struct windowlist));
    wl->window=W_Void2Window(child);
    win=W_Void2Window(parent);
    wl->next=win->children;
    win->children=wl;
}

deleteChild(parent, child)
W_Window parent, child;
{
    struct windowlist **wl;
    struct window *win;
    struct windowlist *temp;

    win=W_Void2Window(parent);
    wl= &(win->children);
    while ((*wl) && (*wl)->window != W_Void2Window(child)) {
	wl= &((*wl)->next);
    }
    temp= *wl;
    *wl= (*wl)->next;
    free((char *) temp);
}

checkGeometry(name, x, y, width, height)
char *name;
int *x, *y, *width, *height;
{
    char *adefault;
    char buf[100];
    char *s;

    sprintf(buf, "%s.geometry", name);
    adefault=getdefault(buf);
    if (adefault==NULL) return;
    /* geometry should be of the form 502x885+1+1, 502x885, or +1+1 */
    s=adefault;
    if (*s!='+') {
        while (*s!='x' && *s!=0) s++;
        *width = atoi(adefault);
        if (*s==0) return;
        s++;
        adefault=s;
        while (*s!='+' && *s!=0) s++;
        *height = atoi(adefault);
        if (*s==0) return;
    }
    s++;
    adefault=s;
    while (*s!='+' && *s!=0) s++;
    *x = atoi(adefault);
    if (*s==0) return;
    s++;
    adefault=s;
    *y = atoi(adefault);
    return;
}

checkParent(name, parent)
char *name;
W_Window *parent;
{
    char *adefault;
    char buf[100];
    int i;
    struct windowlist *windows;
    sprintf(buf, "%s.parent", name);
    adefault=getdefault(buf);
    if (adefault==NULL) return;
    /* parent must be name of other window or "root" */
    if (strcmpi(adefault, "root")==0) {
        *parent=NULL;
        return;
    }
    for (i=0; i<HASHSIZE; i++) {
        windows=hashtable[i];
        while (windows != NULL) {
            if (strcmpi(adefault, windows->window->name)==0) {
                *parent=W_Window2Void(windows->window);
                return;
            }
            windows=windows->next;
        }
    }
}

checkMapped(name)
char *name;
{
    char *adefault;
    char buf[100];

    sprintf(buf, "%s.mapped", name);
    return(booleanDefault(buf, 0));
}

struct window *
findWindow(window)
long window;
{
    struct windowlist *entry;

    entry=hashtable[hash(window)];
    while (entry!=NULL) {
	if (entry->window->window == window) return(entry->window);
	    entry=entry->next;
    }
    return(NULL);
}

addToHash(win)
struct window *win;
{
    struct windowlist **new;

    new= &hashtable[hash(win->window)];
    while (*new != NULL) {
        new= &((*new)->next);
    }
    *new=(struct windowlist *) malloc(sizeof(struct windowlist));
    (*new)->next=NULL;
    (*new)->window=win;
}

removeFromHash(win)
struct window *win;
{
    struct windowlist **temp;
    struct windowlist *old;

    temp= &hashtable[hash(win->window)];
    while ((*temp)->window != win && (*temp) != NULL) {
        temp= &((*temp)->next);
    }
    if ((*temp)->window == win) {
	old=(*temp);
	(*temp)=old->next;
	free((char *) old);
    }
}

shapewindow(win)
struct window *win;
{
    if (win->window == 0) return;
    setWindow(W_Window2Void(win));
    setColor(W_Black);

    clear();
}

setWindow(window)
W_Window window;
{
    struct window *win;

    if (lastwindow == window) return;
    win=W_Void2Window(window);
    winset(win->window);
    lastwindow=window;
}

setColor(acolor)
W_Color acolor;
{
    struct window *win;

    win=W_Void2Window(lastwindow);
    if (win==NULL || win->lastcolor == acolor) return;
    color(colortable[acolor].pixelvalue);
    win->lastcolor=acolor;
}

setFont(fontnum)
int fontnum;
{
    if (lastfont == fontnum) return;
    font(fontnum);
}

drawTile(win)
struct window *win;
{
    if (win->tile == NULL) return;
    if (win->window == 0) return;
    setWindow(W_Window2Void(win));

    rectwrite(win->lb, win->bb,
	win->winwidth-win->rb-1, 
	win->winheight-win->tb-1, win->tile);
}

retile(win,x0,y0,x1,y1)
struct window *win;
int x0,y0,x1,y1;
{
    if (win->tile == NULL) return;
    if (win->window == 0) return;
    setWindow(W_Window2Void(win));
    scrmask(x0+win->lb,x1+win->lb,
	win->winheight-y1-win->tb-2,
	win->winheight-y0-win->tb-2);
    drawTile(win);
    scrmask(win->lb,
        win->winwidth-win->rb-1,
	win->bb,
	win->winheight-win->tb-1);
}

drawBorder(win)
struct window *win;
{
    if (win->window == 0) return;
    setWindow(W_Window2Void(win));
    setColor(win->color);

    viewport(0,win->winwidth-1,0,win->winheight-1);
    ortho2((Coord) 0, (Coord) win->winwidth, (Coord) win->winheight, (Coord) 0);
    
    if (win->tb) rectfi(0,0,win->winwidth,win->tb);
    if (win->lb) rectfi(0,win->tb,win->lb,win->winheight);
    if (win->bb) rectfi(win->lb,win->winheight-win->bb,
	win->winwidth,win->winheight);
    if (win->rb) rectfi(win->winwidth,win->tb,
	win->winwidth-win->rb,win->winheight-win->bb);

    scrmask(win->lb,
        win->winwidth-win->rb-1,
	win->bb,
	win->winheight-win->tb-1);
    ortho2((Coord) -win->lb+win->offx,
        (Coord) win->winwidth+win->offx-win->lb-1,
	(Coord) win->winheight-win->tb+win->offy-1,
	(Coord) -win->tb+win->offy);
}

unsigned short *
x11tocursor(bits)
unsigned char *bits;
{
    unsigned short *cursordata;
    int i,j;

    cursordata=(unsigned short *) malloc(sizeof(short) * 16);
    bzero(cursordata, sizeof(short) * 16);

    for (i=0; i<16; i++) {
	for (j=0; j<16; j++) {
	    if (bits[i*2+j/8] & (1<<(j&7))) {
		cursordata[15-i] |= 1<<(15-j);
	    }
	}
    }
    return(cursordata);
}

loadbits(data, bits, width, height, color)
register short *data;
register unsigned char *bits;
int width, height;
W_Color color;
{
    register int i,j,rot;
    int pixelvalue;

    pixelvalue=colortable[color].pixelvalue;
    data += height*width-width;

    for (i=0; i<height; i++) {
        rot= -1;
        for (j=0; j<width; j++) {
            rot++;
            if (rot==8) {
                rot=0;
                bits++;
            }
            if (*bits & (1<<rot)) {
                *data=pixelvalue;
            }
            data++;
        }
        bits++;
	data -= width*2;
    }
}

unsigned short *
makeTile(icon, width, height)
W_Icon icon;
int width, height;
{
    int i,j,rot,x,y;
    struct icon *data;
    register unsigned short *s;
    unsigned short *tile;
    int whitepixel;
    unsigned char *bits;

    whitepixel=colortable[W_White].pixelvalue;

    tile=(unsigned short *) malloc(width*height*sizeof(unsigned short));
    bzero(tile, width*height*sizeof(unsigned short));
    data=W_Void2Icon(icon);
    bits=data->bits;

    for (i=data->height-1; i>=0; i--) {
        rot= -1;
	for (j=0; j<data->width; j++) {
            rot++;
            if (rot==8) {
                rot=0;
                bits++;
            }
            if (*bits & (1<<rot)) {
		for (x=j; x<width; x+=data->width) {
		    for (y=i; y<height; y+=data->height) {
			tile[y*width+x]=whitepixel;
		    }
		}
            }
        }
        bits++;
    }
    return(tile);
}

#define MAXFONTDATA 131072

Fontchar newfont[256];
unsigned short raster[MAXFONTDATA];
int curbitmapfont=5;
int curbitmapoffset=0;
int curbitmapchar=1;

loadNewBitmap(icon)
struct icon *icon;
{
    int fontwidth;
    int bitswidth;
    int fontsize;
    register int i,j,rot;

    fontwidth=(icon->width+15)/16;
    bitswidth=(icon->width+7)/8;
    fontsize=fontwidth*icon->height;
    if (fontsize + curbitmapoffset > MAXFONTDATA) {
	storeNewFont();
	nextFont();
    }
    icon->charnum=curbitmapchar;
    icon->fontnum=curbitmapfont;
    newfont[curbitmapchar].offset=curbitmapoffset;
    newfont[curbitmapchar].xoff=0;
    newfont[curbitmapchar].yoff=0;
    newfont[curbitmapchar].w=icon->width;
    newfont[curbitmapchar].h=icon->height;
    newfont[curbitmapchar].width=icon->width;

    bzero(raster+curbitmapoffset, fontwidth*icon->height*2);
    for (i=0; i<icon->height; i++) {
        for (j=0; j<icon->width; j++) {
            if (icon->bits[i*bitswidth+j/8] & (1<<(j&7))) {
                raster[curbitmapoffset+(icon->height-i-1)*fontwidth+j/16] |= 1<<(15-(j&15));
            }
        }
    }

    curbitmapoffset += fontsize;
    curbitmapchar++;
    if (curbitmapchar == 256) {
	storeNewFont();
	nextFont();
    }
}

nextFont()
{
    curbitmapfont++;
    curbitmapoffset=0;
    curbitmapchar=1;
}

storeNewFont()
{
    int i;
    int maxheight;

    newfont[0].offset=0;
    newfont[0].w=0;
    newfont[0].h=0;
    newfont[0].xoff=0;
    newfont[0].yoff=0;
    newfont[0].offset=0;
    maxheight=0;
    for (i=1; i<curbitmapchar; i++) { 
	if (newfont[i].h > maxheight) maxheight=newfont[i].h;
    }
    defrasterfont(curbitmapfont, maxheight, curbitmapchar,
	newfont, curbitmapoffset, raster);
    lastFontLoaded=curbitmapfont;
}
