/*
 * interface.c
 *
 * This file will include all the interfaces between the input routines
 *  and the daemon.  They should be useful for writing robots and the
 *  like 
 */
#include "copyright.h"

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "defs.h"
#include "struct.h"
#include "data.h"

set_speed(speed)
int speed;
{
    if (speed > me->p_ship.s_maxspeed) {
	me->p_desspeed = me->p_ship.s_maxspeed;
    } else if (speed < 0) {
	speed=0;
    }
    me->p_desspeed = speed;
    if (me->p_flags & PFDOCK) {
	players[me->p_docked].p_docked--;
	players[me->p_docked].p_port[me->p_port[0]] = VACANT;
    }
    me->p_flags &= ~(PFREPAIR | PFBOMB | PFORBIT | PFDOCK | PFBEAMUP | PFBEAMDOWN);
}

set_course(dir)
unsigned char dir;
{
    me->p_desdir = dir;
    if (me->p_flags & PFDOCK) {
	players[me->p_docked].p_docked--;
	players[me->p_docked].p_port[me->p_port[0]] = VACANT;
    }
    me->p_flags &= ~(PFBOMB | PFORBIT | PFDOCK | PFBEAMUP | PFBEAMDOWN);
}

shield_up()
{
    me->p_flags |= PFSHIELD;
    me->p_flags &= ~(PFBOMB | PFREPAIR | PFBEAMUP | PFBEAMDOWN);
}

shield_down()
{
    me->p_flags &= ~PFSHIELD;
}

shield_tog()
{
    me->p_flags ^= PFSHIELD;
    me->p_flags &= ~(PFBOMB | PFREPAIR | PFBEAMUP | PFBEAMDOWN);
}

bomb_planet()
{
    if (!(me->p_flags & PFORBIT)) {
	warning("Must be orbiting to bomb");
	return;
    }
    me->p_flags |= PFBOMB;
    me->p_flags &= ~(PFSHIELD | PFREPAIR | PFBEAMUP | PFBEAMDOWN);
}

beam_up()
{
    if (!(me->p_flags & (PFORBIT | PFDOCK))) {
	warning("Must be orbiting or docked to beam up.");
	return;
    }
    if (me->p_flags & PFORBIT) {
	if (me->p_team != planets[me->p_planet].pl_owner) {
	    warning("Those aren't our men.");
	    return;
	}
    } else if (me->p_flags & PFDOCK) {
	if (me->p_team != players[me->p_docked].p_team) {
	    warning("Comm Officer: We're not authorized to beam foriegn troops on board!"
);
	    return;
	}
    }
    me->p_flags |= PFBEAMUP;
    me->p_flags &= ~(PFSHIELD | PFREPAIR | PFBOMB | PFBEAMDOWN);
}

beam_down()
{
    if (!(me->p_flags & (PFORBIT | PFDOCK))) {
	warning("Must be orbiting or docked to beam down.");
	return;
    }
    if (me->p_flags & PFDOCK) {
	if (me->p_team != players[me->p_docked].p_team) {
	    warning("Comm Officer: Starbase refuses permission to beam our troops over.");
	    return;
	}
    }
    me->p_flags |= PFBEAMDOWN;
    me->p_flags &= ~(PFSHIELD | PFREPAIR | PFBOMB | PFBEAMUP);
}

repair()
{
    me->p_desspeed = 0;
    me->p_flags |= PFREPAIR;
    me->p_flags &= ~(PFSHIELD | PFBOMB | PFBEAMUP | PFBEAMDOWN | PFPLOCK | PFPLLOCK);
}

repair_off()
{
    me->p_flags &= ~PFREPAIR;
}

repeat_message()
{
    if (++lastm == MAXMESSAGE) ;
	lastm = 0;
}

cloak()
{
    me->p_flags ^= PFCLOAK;
    me->p_flags &= ~(PFTRACT | PFPRESS);
}

cloak_on()
{
    me->p_flags |= PFCLOAK;
    me->p_flags &= ~(PFTRACT | PFPRESS);
}

cloak_off()
{
    me->p_flags &= ~PFCLOAK;
}

lock_planet(planet)
int planet;
{
    char buf[80];
	
    if (planet<0 || planet>=MAXPLANETS) return;

    me->p_flags |= PFPLLOCK;
    me->p_flags &= ~(PFPLOCK|PFORBIT|PFBEAMUP|PFBEAMDOWN|PFBOMB);
    me->p_planet = planet;
    sprintf(buf, "Locking onto %s", planets[planet].pl_name);
    warning(buf);
}

lock_player(player)
int player;
{
    char buf[80];

    if (player<0 || player>=MAXPLAYER) return;
    if (players[player].p_status != PALIVE) return;
    if (players[player].p_flags & PFCLOAK) return;

    me->p_flags |= PFPLOCK;
    me->p_flags &= ~(PFPLLOCK|PFORBIT|PFBEAMUP|PFBEAMDOWN|PFBOMB);
    me->p_playerl = player;
    sprintf(buf, "Locking onto %s (%c%c)",
	players[player].p_name,
	teamlet[players[player].p_team],
	shipnos[players[player].p_no]);
    warning(buf);
}

declare_war(mask)
int mask;
{
    int changes;
    int i;

    mask &= ALLTEAM;
    mask &= ~me->p_team;
    mask |= me->p_swar;
    changes=mask ^ me->p_hostile;
    if (changes==0) return;

    if (changes & FED) {
	sendwarn("Federation", mask & FED, FED);
    }
    if (changes & ROM) {
	sendwarn("Romulans", mask & ROM, ROM);
    }
    if (changes & KLI) {
	sendwarn("Klingons", mask & KLI, KLI);
    }
    if (changes & ORI) {
	sendwarn("Orions", mask & ORI, ORI);
    }
    if (me->p_flags & PFDOCK) {
	if (players[me->p_docked].p_team & mask) {
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	    players[me->p_docked].p_docked--;
	    me->p_flags &= ~PFDOCK;
	}
    } else if (me->p_ship.s_type == STARBASE) {
	if (me->p_docked > 0) {
	    for(i=0; i<NUMPORTS; i++) {
		if (mask & players[me->p_port[i]].p_team) {
		    players[me->p_port[i]].p_flags &= ~PFDOCK;
		    me->p_docked--;
		    me->p_port[i] = VACANT;
		}
	    }
	}
    }

    if (mask & ~me->p_hostile) {
	me->p_flags |= PFWAR;
	delay = me->p_updates + 100;
	warning("Pausing ten seconds to re-program battle computers.");
    }
    me->p_hostile=mask;
}

sendwarn(string, atwar, team)
char *string;
int atwar;
int team;
{
    char buf[BUFSIZ];
    char addrbuf[10];

    if (atwar) {
        (void) sprintf(buf, "%s (%c%c) declaring war on the %s",
            me->p_name,
            teamlet[me->p_team],
            shipnos[me->p_no],
            string);
    }
    else {
        (void) sprintf(buf, "%s (%c%c) declaring peace with the %s",
            me->p_name,
            teamlet[me->p_team],
            shipnos[me->p_no],
            string);
    }

    (void) sprintf(addrbuf, " %c%c->%-3s",
        teamlet[me->p_team],
        shipnos[me->p_no],
        teamshort[team]);
    pmessage2(buf, team, MTEAM, addrbuf, me->p_no);
}

do_refit(type)
     int type;
{	
    int i=0, troop_capacity=0;

    if (type<0 || type>=ATT) return;
    if (me->p_flags & PFORBIT) {
	if (!(planets[me->p_planet].pl_flags & PLHOME)) {
	    warning("You must orbit your HOME planet to apply for command reassignment!");
	    return;
	} else {
	    if (!(planets[me->p_planet].pl_flags & me->p_team)) {
		warning("You must orbit your home planet to apply for command reassignment!");
		return;
	    }
	} /* if (PLHOME) */
    } else if (me->p_flags & PFDOCK) {
	if (type == STARBASE) {
	    warning("Can only refit to starbase on your home planet.");
	    return;
	}
	if (players[me->p_docked].p_team != me->p_team) {
	    warning("You must dock YOUR starbase to apply for command reassignment!");
	    return;
	}
    } else {
	warning("Must orbit home planet or dock your starbase to apply for command reassignment!");
	return;
    } 

    if ((me->p_damage> ((float)me->p_ship.s_maxdamage)*.75) || 
	    (me->p_shield < ((float)me->p_ship.s_maxshield)*.75) ||
	    (me->p_fuel < ((float)me->p_ship.s_maxfuel)*.75)) {
	warning("Central Command refuses to accept a ship in this condition!");
	return;
    }

    if ((me->p_armies > 0)) {
	warning("You must beam your armies down before moving to your new ship");
	return;
    }

    if (shipsallowed[type]==0) {
	warning("That ship hasn't been designed yet.");
	return;
    }

    if (type == STARBASE) {
	if (me->p_stats.st_rank < sbrank) {
	    char buf[100];

	    sprintf(buf, "You need a rank of %s or higher to command a starbase!", ranks[sbrank].name);
	    warning(buf);
	    return;
	}
    }
    if (type == STARBASE && !chaos) {
	for (i=0; i<MAXPLAYER; i++) 
	    if ((me->p_no != i) && 
		(players[i].p_status == PALIVE) && 
		(players[i].p_team == me->p_team))
		if (players[i].p_ship.s_type == STARBASE) {
		    warning("Your side already has a starbase!");
		    return;
		}
    }
    if (type == STARBASE && !chaos) {
	if (numShips(me->p_team) < 4) {
	    warning("Your team is not capable of defending such an expensive ship");
	    return;
	}
    }
    if (type == STARBASE && !chaos) {
	if (teams[me->p_team].s_turns > 0) {
	    warning("Your new starbase is still under construction");
	    return;
	}
    }
    if (type == STARBASE && numPlanets(me->p_team) < 5 && !chaos) {
	warning("Your team's stuggling economy cannot support such an expenditure!");
	return;
    }

    if ((me->p_ship.s_type == STARBASE) && (type != STARBASE)) {
	/* Reset kills to 0.0 */
	me->p_kills=0;
	/* bump all docked ships */
	for (i=0; i<NUMPORTS; i++) 
	   if (me->p_port[i] != VACANT) {
	       players[me->p_port[i]].p_flags &= ~PFDOCK;
	       me->p_docked--;
	       me->p_port[i] = VACANT;
	   }
    }	

    
    getship(&(me->p_ship), type, me->p_team);
    if (type != STARBASE && me->p_kills < plkills) {
	me->p_ship.s_plasmacost = -1;
    }
    me->p_shield = me->p_ship.s_maxshield;
    me->p_damage = 0;
    me->p_fuel = me->p_ship.s_maxfuel;
    me->p_wtemp = 0;
    me->p_etemp = 0;
    if (type == STARBASE) {
	me->p_docked = 0;
	for (i=0; i<4; i++) me->p_port[i] = VACANT;
    }

    me->p_flags &= ~(PFREFIT);
    me->p_flags |= PFREFITTING;
    rdelay = me->p_updates + 50;

    warning("You are being transported to your new vessel .... ");
}

numShips(owner)
int owner;
{
    int i, num=0;
    struct player *p;

    for (i=0, p=players; i<MAXPLAYER; i++, p++) {
        if (p->p_status==PALIVE && p->p_team==owner) {
            num++;
        }
    }
    return (num);
}

numPlanets(owner)
int owner;
{
    int i, num=0;
    struct planet *p;

    for (i=0, p=planets; i<MAXPLANETS; i++, p++) {
        if (p->pl_owner==owner)  {
            num++;
        }
    }
    return(num);
}

