/*
 * v1.0 by ksmith & scott (presumably).
 *
 * v1.1 by Andy McFadden.
 * - Displays player/planet stats continuously.
 * - Updates messages while in player/planet display.
 * - Displays "daemon not running" after the daemon shuts down.
 * - Displays "nobody is playing" if nobody is in the game.
 * - Doesn't lock up under UTS when typing message text.
 * - Panic button clears the screen and waits for a key (for fascist environs).
 * - Aware of ATT-class starships.
 *
 * 02-Dec-91
 */
#include <stdio.h>
/*#include <strings.h>*/
#include <string.h>		/* ATM */
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/file.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include <curses.h>
#include "defs.h"
#include "struct.h"

struct player *players;
struct player *me;
struct torp *torps;
struct ship *myship;
struct stats *mystats;
struct planet *planets;
struct phaser *phasers;
struct message *messages;
struct mctl *mctl;
struct team *teams;
struct plasmatorp *plasmatorps;
struct status *status;

#define HEIGHT 20
#define WIDTH 80
#define MESSAGES 3
#define THEIGHT HEIGHT+MESSAGES
char *ships = "0123456789abcdefghij";
char *teamnames[] = {"---","FED","ROM","???","KLI","???","???","???","ORI"};

struct memory	*sharedMemory;
int delay=5;		/* ATM: was 10 (was local, too) */
int statdelay=10;	/* delay for planet/player displays */
int messageline,	/* was local */
    oldmctl;		/* was local */

/*
 * ATM: Waits until the daemon is up and people are playing.
 */
openmem()
{
    extern int errno;
    int	shmemKey = PKEY;
    int	shmid;
/*    struct memory	*sharedMemory;*/
    int i, state=-1;

    while (1) {
	errno = 0;
	shmid = shmget(shmemKey, 0, 0);
	if (shmid < 0) {
	    if (errno != ENOENT) {
		fprintf(stderr, "shmget\n");
		exit(1);
	    }
	    shmid = shmget(shmemKey, 0, 0);
	    if (shmid < 0) {
		if (state != 0) {
		    printf("\r  %-30s\r", "Daemon not running.");
		    fflush(stdout);
		    state = 0;
		}
		sleep(20);
		continue;
	    }
	}
	sharedMemory = (struct memory *) shmat(shmid, 0, 0);
	if (sharedMemory == (struct memory *) -1) {
	    fprintf(stderr, "shared memory\n");
	    exit(1);
	}
	players = sharedMemory->players;
	for (i = 0; i < MAXPLAYER; i++) {
	    if (players[i].p_status != PFREE && players[i].p_stats.st_tticks!=0)
		goto okay;
	}

	/* nobody playing, remove the segment, sleep, and try again */
	if (state != 1) {
	    printf("\r  %-30s\r", "Nobody is playing.");  fflush(stdout);
	    state = 1;
	}
	if (shmdt(sharedMemory) < 0) {
	    perror("shmdt");
	    exit(1);
	}
	sleep(20);
    }

okay:
    printf("\r  %-30s\r", " ");  fflush(stdout);	/* erase the line */
    players = sharedMemory->players;
    torps = sharedMemory->torps;
    plasmatorps = sharedMemory->plasmatorps;
    planets = sharedMemory->planets;
    phasers = sharedMemory->phasers;
    mctl = sharedMemory->mctl;
    messages = sharedMemory->messages;
    teams = sharedMemory->teams;
    status = sharedMemory->status;
}

reaper()
{
    clear();
    refresh();
    endwin();
    exit(0);
}

WINDOW *mesg_w, *main_w, *command_w;
char lc[]="ifrxkxxxo";

main(argc, argv)
int argc;
char **argv;
{
    int i,j,k;
    char s[80];
    int command=0;
    int zoom= -1;
    char *st;
    int sendingMessage=0;
    int sendingTo=0;
    int sendMask=0;
    char header[20];

    for (i=1; i<argc; i++) {
	if (argv[i][0]=='-') {
	    switch(argv[i][1]) {
	    case 'd':
		delay=atoi(argv[i]+2);
		if (delay==0) delay=10;
		break;
	    default:
		usage(argv[0]);
		break;
	    }
	}
    }

    /* curses init */
    initscr();
    signal(SIGINT, reaper);
/*    signal(SIGTSTP, reaper);*/	/* I'm lazy, ok? */	/* ATM */
    cbreak();				/* ATM: was crmode();*/
    noecho();
    main_w=newwin(HEIGHT, WIDTH, 0, 0);
    mesg_w=newwin(MESSAGES, WIDTH, HEIGHT, 0);
    command_w=newwin(1, WIDTH, THEIGHT, 0);
    wrefresh(mesg_w);
    wrefresh(command_w);
    wrefresh(main_w);

restart:
    openmem();		/* doesn't return until people are playing */
    oldmctl=mctl->mc_current;
    command=0;
    zoom= -1;
    sendingMessage=0;
    sendingTo=0;
    sendMask=0;

    messageline=0;
    while (1) {
	/* if nobody is playing, restart and wait */
	for (i = 0; i < MAXPLAYER; i++)
	    if (players[i].p_status != PFREE && players[i].p_stats.st_tticks!=0)
		goto okay;
	if (shmdt(sharedMemory) < 0) {		/* kill segment & restart */
	    perror("shmdt");
	    exit(1);
	}
	wclear(mesg_w);		wrefresh(mesg_w);
	wclear(command_w);	wrefresh(command_w);
	wclear(main_w);		wrefresh(main_w);
	goto restart;

okay:
	nodelay(command_w, 1);			/* ATM */
	while (isinput(delay)) {
	    int c;
	    if ((c = wgetch(command_w)) == ERR)
		break;		/* ATM: equivalent to !isinput() */
	    ungetch(c);
	    if (!command) {
		command=getch();
		if (sendingMessage) {
		    if (sendingTo==-1) {
			strcpy(header, "GOD->");
			switch(command) {
			case 27:
			    sendingMessage=0;
			    wclear(command_w);
			    wrefresh(command_w);
			    break;
			case 'A':
			    sendingTo=0;
			    sendMask=MALL;
			    strcat(header,"ALL ");
			    break;
			case 'F':
			    sendingTo=FED;
			    sendMask=MTEAM;
			    strcat(header,"FED ");
			    break;
			case 'R':
			    sendingTo=ROM;
			    sendMask=MTEAM;
			    strcat(header,"ROM ");
			    break;
			case 'K':
			    sendingTo=KLI;
			    sendMask=MTEAM;
			    strcat(header,"KLI ");
			    break;
			case 'O':
			    sendingTo=ORI;
			    sendMask=MTEAM;
			    strcat(header,"ORI ");
			    break;
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
			    sendingTo=command-'0';
			    sendMask=MINDIV;
			    break;
			case 'a': case 'b': case 'c': case 'd': case 'e':
			case 'f': case 'g': case 'h': case 'i': case 'j':
			    sendingTo=command-'a'+10;
			    sendMask=MINDIV;
			    break;
			default:
			    break;
			}
			if (sendMask==MINDIV) {
			    strncat(header, players[sendingTo].p_mapchars, 2);
			    strcat(header, "  ");
			}
			if (sendingTo != -1) {
			    wmove(command_w, 0,0);
			    waddstr(command_w, ": ");
			    waddstr(command_w, header);
			    waddstr(command_w, "_");
			}
		    } else {
			static char theMess[80];
			static char *s=NULL;

			if (s==NULL) {
			    strcpy(theMess,header);
			    s=theMess+strlen(theMess);
			}
			switch(command) {
			case 27:	/* Escape */
			    s=NULL;
			    sendingMessage=0;
			    wclear(command_w);
			    break;
			case '\n':
			    pmessage(theMess,sendingTo,sendMask);
			    s=NULL;
			    sendingMessage=0;
			    wclear(command_w);
			    break;
			case '\b':
			case 127:	/* Delete */
			    if (s>theMess) {
				s--;
				*s=0;
				wmove(command_w,0,s-theMess+2);
				waddstr(command_w,"_ ");
				break;
			    }
			    break;
			case 21:	/* Ctrl - U */
			    s=theMess;
			    wclear(command_w);
			    wmove(command_w,0,0);
			    waddstr(command_w,": _");
			    break;
			default:
			    if (s<theMess+76) {
				*s=command;
				s++;
				*s=0;
				wmove(command_w,0,s-theMess+1);
				waddch(command_w,command);
				waddch(command_w,'_');
				break;
			    }
			    break;
			}
		    }
		    wrefresh(command_w);
		} else switch(command) {
		case 'm':
		    sendingMessage=1;
		    sendingTo= -1;
		    wmove(command_w,0,0);
		    waddstr(command_w,"?");
		    wrefresh(command_w);
		    break;
		case 'f':
		    zoom= -1;
		    command=0;
		    break;
		case 'z':
		    break;
		case 12:
		    command=0;
		    wclear(mesg_w);
		    wclear(main_w);
		    wclear(command_w);
		    wrefresh(mesg_w);
		    wrefresh(main_w);
		    wrefresh(command_w);
		    printf("\033[;H\033[2J");
		    fflush(stdout);
		    break;
		case 'P':
		    showPlanets();
		    wclear(main_w);
		    break;
		case 'L':
		    showPlayers();
		    wclear(main_w);
		    break;
		case '.':		/* "panic" mode */
		    wclear(mesg_w);	/* clear them so they get redrawn */
		    wclear(main_w);
		    wclear(command_w);
		    printf("\033[H\033[2J");	/* vt100 "erase screen" */
		    fflush(stdout);
		    while (1) {
			sleep (1);
			if ((c = getch()) != ERR)
			break;
		    }
		    command=0;
		    break;
		default:
		    command=0;
		    break;
		}
	    } else {
		switch(command) {
		case 'z':
		    st=index(ships,getch());
		    if (st!=0) {
			zoom=st-ships;
		    }
		    break;
		default:
		    break;
		}
		command=0;
	    }
	}
#ifdef OLDM
	if (oldmctl!=mctl->mc_current) {
	    oldmctl++;
	    if (oldmctl==MAXMESSAGE) oldmctl=0;
	    wmove(mesg_w, messageline, 0);
	    waddstr(mesg_w, " ");
	    messageline++;
	    if (messageline==MESSAGES) messageline=0;
	    wmove(mesg_w, messageline, 0);
	    waddstr(mesg_w, ">");
	    wclrtoeol(mesg_w);
	    waddstr(mesg_w, messages[oldmctl].m_data);
	}
#else
	update_msgs();
#endif
	werase(main_w);
	for (i=0; i<MAXPLANETS; i++) {
	    strncpy(s+2,planets[i].pl_name,3);
	    s[5]=0;
	    s[1]='-';
	    s[0]=teamnames[planets[i].pl_owner][0];
	    writescreen(zoom, planets[i].pl_x,planets[i].pl_y,s,5);
	}
	for (i=0; i<MAXPLAYER; i++) {
	    if (players[i].p_status != PFREE) {
	    }
	    if (players[i].p_status != PALIVE) continue;
	    strncpy(s, players[i].p_mapchars,2);
	    s[2]=0;
	    if ((players[i].p_flags & PFSEEN) == 0) {
		if (players[i].p_flags & PFCLOAK) {
		    s[0]='-';
		} else {
		    s[0]=lc[players[i].p_team];
		}
	    } else if (players[i].p_flags & PFCLOAK) {
		s[0]='?';
	    }
	    writescreen(zoom, players[i].p_x, players[i].p_y, 
		s, 2);
	}
	for (i=0; i<MAXPLAYER*MAXTORP; i++) {
	    if (torps[i].t_status==TFREE) continue;
	    if (torps[i].t_status==TEXPLODE) {
		writescreen(zoom, torps[i].t_x, torps[i].t_y, "o", 1);
	    } else {
		writescreen(zoom, torps[i].t_x, torps[i].t_y, ".", 1);
	    }
	}
	for (i=0; i<MAXPLAYER*MAXPLASMA; i++) {
	    if (plasmatorps[i].pt_status==PTFREE) continue;
	    if (plasmatorps[i].pt_status==PTEXPLODE) {
		writescreen(zoom, plasmatorps[i].pt_x, plasmatorps[i].pt_y, 
		    "*", 1);
	    } else {
		writescreen(zoom, plasmatorps[i].pt_x, plasmatorps[i].pt_y, 
		    "O", 1);
	    }
	}
	wrefresh(main_w);
/*	wrefresh(mesg_w);*/
	wrefresh(command_w);
    }
}

update_msgs()
{
    if (oldmctl!=mctl->mc_current) {
	oldmctl++;
	if (oldmctl==MAXMESSAGE) oldmctl=0;
	wmove(mesg_w, messageline, 0);
	waddstr(mesg_w, " ");
	messageline++;
	if (messageline==MESSAGES) messageline=0;
	wmove(mesg_w, messageline, 0);
	waddstr(mesg_w, ">");
	wclrtoeol(mesg_w);
	waddstr(mesg_w, messages[oldmctl].m_data);

	if (!strncmp(messages[oldmctl].m_data +5, "GOD", 3))
	    beep();

	wrefresh(mesg_w);
    }
}

writescreen(zoom,x,y,str,len)
int zoom;
int x,y,len;
char *str;
{
    int j,k;
    int i;

    if (zoom== -1) {
	if (x==GWIDTH) x=GWIDTH-1;
	if (y==GWIDTH) y=GWIDTH-1;
	j=(WIDTH)*x/GWIDTH;
	k=(HEIGHT)*y/GWIDTH;
    } else {
	j=(WIDTH)*(x-players[zoom].p_x)/(WINSIDE*SCALE)+WIDTH/2;
	k=(HEIGHT)*(y-players[zoom].p_y)/(WINSIDE*SCALE)+HEIGHT/2;
    }
    j=j-len/2;
    for (i=0; i<len; i++) {
	writechar(j+i,k,str[i]);
    }
}

writechar(x,y,ch)
int x,y;
char ch;
{
    if (x>=0 && x<WIDTH && y>=0 && y<HEIGHT) {
	wmove(main_w, y,x);
	waddch(main_w, ch);
    }
}

isinput(del)
int del;
{
    struct timeval Utimeout;		/* ATM: "timeout" wouldn't compile */
    fd_set readfds, writefds, exceptfds;

    FD_ZERO(&readfds);  FD_ZERO(&writefds);  FD_ZERO(&exceptfds);
    FD_SET(fileno(stdin), &readfds);
    Utimeout.tv_sec = del/10;
    Utimeout.tv_usec = (del % 10) * 100000;
    /*
     * select() and curses don't seem to get along, so we just use select()
     * for the delay.  Otherwise, select() can hang the program if you type
     * too quickly.  (It's supposed to return immediately if a key is waiting,
     * which it does.  Sometimes.)  The result of this change is that
     * everything appears normal until you try to send a message; in that case,
     * it hangs (under UTS) until you stop typing for a full delay period.
     * Your mileage may vary, especially if your select() isn't brain-damaged.
     * ATM
     */
    (void) select(1, &readfds, &writefds, &exceptfds, &Utimeout);
    return(1);
}

showPlanets()
{
    int i,j,k;

    wclear(main_w);
    while (1) {
	j=0; 
	k=0;
	for (i=0; i<MAXPLANETS; i++) {
	    wmove(main_w,j,k);
	    wprintw(main_w,"%-16s %s %4d  %c%c%c  %c%c%c%c",
		planets[i].pl_name, teamnames[planets[i].pl_owner],
		planets[i].pl_armies, 
		(planets[i].pl_flags & PLFUEL) ? 'F' : ' ',
		(planets[i].pl_flags & PLREPAIR) ? 'R' : ' ',
		(planets[i].pl_flags & PLAGRI) ? 'A' : ' ',
		(planets[i].pl_info & FED) ? 'F' : ' ',
		(planets[i].pl_info & ROM) ? 'R' : ' ',
		(planets[i].pl_info & KLI) ? 'K' : ' ',
		(planets[i].pl_info & ORI) ? 'O' : ' ');
	    j=j+1;
	    if (j==HEIGHT) {
		j=0;
		k=k+40;
		if (k>WIDTH-30) {
		    break;
		}
	    }
	}
	wrefresh(main_w);
	update_msgs();
	(void) isinput(statdelay);	/* wait for a bit */
	if (wgetch(main_w) != ERR)	/* check for char */
	    break;
    }
}

char *shiptypes[] = {"SC", "DD", "CA", "BB", "AS", "SB", "??"};

showPlayers()
{
    int i,j;
    static int mode=0;
    int c;

    if (mode) goto mode1;

    for (;;) {
	mode = 0;
	wclear(main_w);
	while (1) {
	    wmove(main_w,0,0);
	    wprintw(main_w," # Player          Display          Type Bombing Planets Offense Defense");
	    for (i=0; i<MAXPLAYER; i++) {
		wmove(main_w,i+1,0);
		if (players[i].p_status==PFREE) {
		    wclrtoeol(main_w);
		    continue;
		}
		wprintw(main_w,"%2s %-16s%-16.16s   %2s %7.2f %7.2f %7.2f %7.2f",
		    players[i].p_mapchars,
		    players[i].p_name,
		    players[i].p_monitor,
		    shiptypes[players[i].p_ship.s_type],
		    bombingRating(players+i),
		    planetRating(players+i),
		    offenseRating(players+i),
		    defenseRating(players+i));
		if (i==HEIGHT) break;		/* shouldn't happen */
	    }
	    wrefresh(main_w);
	    update_msgs();
	    (void) isinput(statdelay);			/* pause */
	    if ((c = wgetch(main_w)) == ' ') break;	/* toggle display */
	    if (c != ERR) return;
	}  /* while(1) */

mode1:
	mode = 1;
	wclear(main_w);
	while (1) {
	    wmove(main_w,0,0);
	    wprintw(main_w," # Player          Login            Type  Kills Damage Shields Armies  Fuel");
	    for (i=0; i<MAXPLAYER; i++) {
		wmove(main_w,i+1,0);
		if (players[i].p_status==PFREE) {
		    wclrtoeol(main_w);
		    continue;
		}
		wprintw(main_w,"%2s %-16s%-16.16s   %2s %6.2f %6d %7d %6d %5d",
		    players[i].p_mapchars,
		    players[i].p_name,
		    players[i].p_login,
		    shiptypes[players[i].p_ship.s_type],
		    players[i].p_kills,
		    players[i].p_damage,
		    players[i].p_shield,
		    players[i].p_armies,
		    players[i].p_fuel);
		if (i==HEIGHT) break;
	    }
	    wrefresh(main_w);
	    update_msgs();
	    (void) isinput(statdelay);			/* pause */
	    if ((c = wgetch(main_w)) == ' ') break;	/* toggle display */
	    if (c != ERR) return;
	}  /* while(1) */

    }  /* for(;;) */
}

usage(string)
char *string;
{
    printf("Usage: %s [-dnnn]\n", string);
    printf("  -dnnn  delay nnn 1/10 seconds between frames\n");
}

pmessage(str, recip, group)
char *str;
int recip;
int group;
{
    struct message *cur;
    if (++(mctl->mc_current) >= MAXMESSAGE)
	mctl->mc_current = 0;
    cur = &messages[mctl->mc_current];
    cur->m_no = mctl->mc_current;
    cur->m_flags = group;
    cur->m_time = 0;
    cur->m_recpt = recip;
    cur->m_from = 19;
    (void) sprintf(cur->m_data, "%s", str);
    cur->m_flags |= MVALID;
}
